""" Utility functions and data.

The functions in the module work directly on data and metadata. In order to
make this easy to write and test, functionality that requires direct
interaction with HDF5 are not included here.

"""

from datetime import datetime
import re
import string
import time

import numpy as np

from .exceptions import BadSDAFile


# DD-MMM-YYYY HH:MM:SS
# MATLAB code uses 'datestr' to create this. HH:MM:SS is optional if all zero.
# This variant is DATE_FORMAT_SHORT.
DATE_FORMAT = "%d-%b-%Y %H:%M:%S"
DATE_FORMAT_SHORT = "%d-%b-%Y"

# Record groups.
SIMPLE_RECORD_TYPES = ('character', 'logical', 'numeric', 'file')
SUPPORTED_RECORD_TYPES = (
    'character', 'file', 'logical', 'numeric', 'cell', 'structure',
    'structures', 'object', 'objects',
)


# Regular expression for version string
VERSION_1_RE = re.compile(r'1\.(?P<sub>\d+)')


# Type codes for unsupported numeric types
UNSUPPORTED_NUMERIC_TYPE_CODES = {
    'G',  # complex256
    'g',  # float128
    'e',  # float16
}

# Empty values for supported types
EMPTY_FOR_TYPE = {
    'numeric': np.nan,
    'character': '',
    'file': b'',
    'logical': np.array([], dtype=bool),
    'cell': [],
    'structure': {}
}

# Equivalent record types for reading
STRUCTURE_EQUIVALENT = {'structure', 'object'}
CELL_EQUIVALENT = {'cell', 'objects', 'structures'}


# Cell label template and generator function
CELL_LABEL_TEMPLATE = "element {}"
cell_label = CELL_LABEL_TEMPLATE.format


def are_record_types_equivalent(rt1, rt2):
    """ Determine if record types are equivalent with respect to reading """
    if rt1 == rt2:
        return True

    if rt1 in STRUCTURE_EQUIVALENT and rt2 in STRUCTURE_EQUIVALENT:
        return True

    if rt1 in CELL_EQUIVALENT and rt2 in CELL_EQUIVALENT:
        return True

    return False


def are_signatures_equivalent(sig1, sig2):
    """ Verify if data signatures are equivalent.

    Parameters
    ----------
    sig1, sig2 :
        Data or group signatures returned by unnest or unnest_record.

    """
    if len(sig1) != len(sig2):
        return False

    for item1, item2 in zip(sig1, sig2):
        key1, rt1 = item1
        key2, rt2 = item2
        if key1 != key2:
            return False

        if not are_record_types_equivalent(rt1, rt2):
            return False

    return True


def error_if_bad_attr(h5file, attr, is_valid):
    """ Raise BadSDAFile error if h5file has a bad SDA attribute.

    This assumes that the attr is stored as bytes. The passed ``is_valid``
    function should accept the value as a string.

    """
    name = h5file.filename
    try:
        value = h5file.attrs[attr]
    except KeyError:
        msg = "File '{}' does not contain '{}' attribute".format(name, attr)
        raise BadSDAFile(msg)
    else:
        value = value.decode('ascii')
        if not is_valid(value):
            msg = "File '{}' has invalid '{}' attribute".format(name, attr)
            raise BadSDAFile(msg)


def error_if_bad_header(h5file):
    """ Raise BadSDAFile if SDA header attributes are missing or invalid. """
    # FileFormat flag
    error_if_bad_attr(h5file, 'FileFormat', is_valid_file_format)

    # FormatVersion flag
    error_if_bad_attr(h5file, 'FormatVersion', is_valid_format_version)

    # Writable flag
    error_if_bad_attr(h5file, 'Writable', is_valid_writable)

    # Created flag
    error_if_bad_attr(h5file, 'Created', is_valid_date)

    # Updated flag
    error_if_bad_attr(h5file, 'Updated', is_valid_date)


def error_if_not_writable(h5file):
    """ Raise an IOError if an SDAFile indicates 'Writable' as 'no'. """
    writable = h5file.attrs.get('Writable')
    if writable == b'no':
        msg = "File '{}' is not writable".format(h5file.filename)
        raise IOError(msg)


def get_date_str(dt=None):
    """ Get a valid date string from a datetime, or current time. """
    if dt is None:
        dt = datetime.now()
    if dt.hour == dt.minute == dt.second == 0:
        fmt = DATE_FORMAT_SHORT
    else:
        fmt = DATE_FORMAT
    date_str = dt.strftime(fmt)
    return date_str


def get_empty_for_type(record_type):
    """ Get the empty value for a record.

    Raises
    ------
    ValueError if ``record_type`` does not have an empty entry.

    """
    try:
        return EMPTY_FOR_TYPE[record_type]
    except KeyError:
        msg = "Record type '{}' cannot be empty".format(record_type)
        raise ValueError(msg)


def is_simple(record_type):
    """ Check if record type is simple (primitive or 'file'). """
    return record_type in SIMPLE_RECORD_TYPES


def is_supported(record_type):
    """ Check if record type is supported. """
    return record_type in SUPPORTED_RECORD_TYPES


def is_valid_date(date_str):
    """ Check date str conforms to DATE_FORMAT or DATE_FORMAT_SHORT. """
    try:
        time.strptime(date_str, DATE_FORMAT)
    except ValueError:
        try:
            time.strptime(date_str, DATE_FORMAT_SHORT)
        except ValueError:
            return False
    return True


def is_valid_file_format(value):
    """ Check that file format is equivalent to 'SDA' """
    return value == 'SDA'


def is_valid_format_version(value):
    """ Check that version is '1.X' for X <= 1 """
    m = VERSION_1_RE.match(value)
    if m is None:
        return False
    return 0 <= int(m.group('sub')) <= 1


def is_valid_matlab_field_label(label):
    """ Check that passed string is a valid MATLAB field label """
    if not label.startswith(tuple(string.ascii_letters)):
        return False
    VALID_CHARS = set(string.ascii_letters + string.digits + "_")
    return set(label).issubset(VALID_CHARS)


def is_valid_writable(value):
    """ Check that writable flag is 'yes' or 'no' """
    return value == 'yes' or value == 'no'


def set_encoded(dict_like, **attrs):
    """ Encode and insert values into a dict-like object. """
    encoded = {
        attr: value.encode('ascii') if isinstance(value, str) else value
        for attr, value in attrs.items()
    }
    dict_like.update(encoded)


def get_decoded(dict_like, *attrs):
    """ Retrieve decoded values from a dict-like object if they exist.

    If no attrs are passed, all values are retrieved.

    """
    # Filter for existing
    if len(attrs) == 0:
        items = dict_like.items()
    else:
        items = [
            (attr, dict_like[attr]) for attr in attrs if attr in dict_like
        ]
    return {
        attr: value.decode('ascii') if isinstance(value, bytes) else value
        for attr, value in items
    }


def get_record_type(dict_like):
    """ Retrived decoded record type from a dict-like object. """
    return get_decoded(dict_like, 'RecordType').get('RecordType')


def unnest(data, registry):
    """ Unnest data.

    Parameters
    ----------
    data :
        Data to unnest
    registry : InserterRegistry
        Registry used to look up inserters

    Returns
    -------
    record_signature :
        A tuple of (path, record_type) tuples where each component of the data
        is identified in the path.

    """
    record_type = getattr(registry.get_inserter(data), 'record_type', None)
    items = [('', record_type, data)]
    for parent, record_type, obj in items:
        if record_type in SIMPLE_RECORD_TYPES:
            continue
        if are_record_types_equivalent(record_type, 'structure'):
            sub_items = sorted(obj.items())
        elif are_record_types_equivalent(record_type, 'cell'):
            sub_items = [
                (cell_label(i), sub_obj)
                for i, sub_obj in enumerate(obj, start=1)
            ]
        for key, sub_obj in sub_items:
            path = "/".join((parent, key)).lstrip("/")
            sub_record_type = getattr(
                registry.get_inserter(sub_obj), 'record_type', None
            )
            items.append((path, sub_record_type, sub_obj))
    return tuple(item[:2] for item in items)


def unnest_record(grp):
    """ Unnest a record group stored on file.

    Parameters
    ----------
    grp : h5py.Group
        The group to unnest.

    Returns
    -------
    record_signature :
        A tuple of (path, record_type) tuples where each component of the group
        is identified in the path.

    """
    record_type = get_record_type(grp.attrs)
    items = [('', record_type, grp)]
    for parent, record_type, obj in items:
        if record_type not in SIMPLE_RECORD_TYPES:
            for key in sorted(obj.keys()):
                path = "/".join((parent, key)).lstrip("/")
                sub_obj = obj[key]
                sub_record_type = get_record_type(sub_obj.attrs)
                items.append((path, sub_record_type, sub_obj))
    return tuple(item[:2] for item in items)


def validate_structures(data, registry):
    """ Validate that passed data is suitable to be a 'structures' record. """
    # Signatures must be homogeneous.
    signatures = set(
        unnest(sub_data, registry) for sub_data in np.ravel(data)
    )
    if len(signatures) > 1:
        msg = "Data cells are not homogenous"
        raise ValueError(msg)

    # The top-most record must be a structure record
    sig = signatures.pop()
    record_type = sig[0][1]
    if record_type != 'structure':
        msg = "Data does not contain structure records"
        raise ValueError(msg)

    return sig


def update_header(attrs):
    """ Update timestamp and version to 1.1 in a header. """
    set_encoded(
        attrs,
        FormatVersion='1.1',
        Updated=get_date_str(),
    )


def write_header(attrs):
    """ Write default, encoded header values to dict-like ``attrs``. """
    date_str = get_date_str()
    set_encoded(
        attrs,
        FileFormat='SDA',
        FormatVersion='1.1',
        Writable='yes',
        Created=date_str,
        Updated=date_str,
    )



from .record_inserter import CompositeRecordInserter, inserter
from .utils import is_valid_matlab_field_label, set_encoded


@inserter
class DictInserter(CompositeRecordInserter):
    """ Inserter for dicts. """

    record_type = 'structure'

    @staticmethod
    def can_insert(data):
        """ This can insert any dict, regardless of the contents. """
        return isinstance(data, dict)

    def record_group_attributes(self, dict_like):
        """ Record FieldNames along with other group attributes. """
        set_encoded(
            dict_like,
            RecordType=self.record_type,
            Empty=self.empty,
            Deflate=self.deflate,
            FieldNames=self.field_names
        )

    def prepare_data(self):
        """ Records FieldNames and Empty metadata, validates keys. """
        self.empty = 'yes' if len(self.data) == 0 else 'no'
        self._keys = sorted(key for key in self.data.keys())

        # Validate the keys
        for key in self._keys:
            if not is_valid_matlab_field_label(key):
                msg = "'{}' is not a valid MATLAB field label".format(key)
                raise ValueError(msg)
        self.field_names = " ".join(self._keys)

    def __iter__(self):
        for key in self._keys:
            sub_data = self.data[key]
            cls = self.registry.get_inserter(sub_data)
            if cls is None:
                msg = "Data not supported for insertion: {!r}".format(sub_data)
                raise ValueError(msg)
            inserter = cls(
                key, sub_data, self.deflate, registry=self.registry
            )
            yield inserter



""" Base class for records. """

from abc import ABCMeta, abstractmethod

from .utils import set_encoded


class InserterRegistry(object):
    """ Registry of inserters. """

    def __init__(self):
        self._inserters = []
        self._register_inserters()

    def _register_inserters(self):
        """ Register all inserters. """
        from itertools import chain

        import sdafile.numeric_inserter as numeric
        import sdafile.logical_inserter as logical
        import sdafile.character_inserter as character
        import sdafile.cell_inserter as cell
        import sdafile.structure_inserter as structure
        import sdafile.file_inserter as file_

        objs = chain(
            numeric.__dict__.values(),
            logical.__dict__.values(),
            character.__dict__.values(),
            cell.__dict__.values(),
            structure.__dict__.values(),
            file_.__dict__.values(),
        )

        inserters = []
        for obj in objs:
            if getattr(obj, '__inserter__', False):
                inserters.append(obj)
        self._inserters = inserters

    def get_inserter(self, data):
        """" Get the inserter appropriate for the passed data.

        This loops through available inserters and uses the first one it
        encounters that can insert the data.

        Parameters
        ----------
        data :
            Data to be inserted into an archive

        Returns
        -------
        inserter : RecordInserter or None
            The RecordInserter *class* that can insert the data into an
            archive, or None if no such inserter can be found.

        """
        for cls in self._inserters:
            if cls.can_insert(data):
                return cls
        return None


def inserter(cls):
    """ Mark a class as an inserter. """
    cls.__inserter__ = True
    return cls


class RecordInserter(object):
    """ Stores a record for insertion. """

    __metaclass__ = ABCMeta

    # The record type supported by the inserter
    record_type = None

    def __init__(self, label, data, deflate, registry=None):
        self.label = label
        self.deflate = int(deflate)
        self.data = self.original_data = data
        self.empty = 'no'
        self._registry = registry

    @property
    def registry(self):
        if self._registry is None:
            self._registry = InserterRegistry()
        return self._registry

    @staticmethod
    @abstractmethod
    def can_insert(data):
        """ Indicates if the Record can insert the passed data.

        This is to be overloaded by derived classes

        """
        return False

    @abstractmethod
    def prepare_data(data):
        """ Prepare data for writing and record metadata.

        This is to be overloaded by derived classes. This is reponsible for
        recording metadata to be written by ``record_group_attributes`` and
        ``record_dataset_attributes``.

        """
        return

    def record_group_attributes(self, dict_like):
        """ Record group attributes specific to the data.

        This includes all group-level data except 'Description'.

        """
        set_encoded(
            dict_like,
            RecordType=self.record_type,
            Empty=self.empty,
            Deflate=self.deflate,
        )

    @abstractmethod
    def insert(self, h5file, description):
        """ Insert the data into an h5py File. """
        return

    @abstractmethod
    def insert_into_group(self, group):
        """ Insert data at the group level """
        return


class SimpleRecordInserter(RecordInserter):
    """ RecordInserter for simple objects.

    Subclasses must convert ``data`` to an ndarray.

    """

    def insert(self, h5file, description):
        """ Insert the data into an h5py File. """
        group = h5file.create_group(self.label)
        set_encoded(
            group.attrs,
            Description=description,
        )
        self.insert_into_group(group)

    def insert_into_group(self, group):
        """ Insert at the group level """
        self.prepare_data()
        self.record_group_attributes(group.attrs)
        self.insert_below_group(group)

    def insert_below_group(self, group):
        """ Insert below a group, creating the necessary dataset entry. """
        maxshape = (None,) * self.data.ndim
        ds = group.create_dataset(
            self.label,
            maxshape=maxshape,
            data=self.data,
            compression=self.deflate,
        )
        self.record_dataset_attributes(ds.attrs)

    def record_dataset_attributes(self, dict_like):
        """ Record the dataset attributes specific to the data. """
        set_encoded(
            dict_like,
            RecordType=self.record_type,
            Empty=self.empty,
        )


class CompositeRecordInserter(RecordInserter):
    """ RecordInserter for composite objects. """

    @abstractmethod
    def __iter__(self):
        """ Yield RecordInserter instances for subitems. """
        return

    def record_dataset_attributes(self, dataset_attrs):
        """ CompositeRecordInserters do not record to datasets. """
        return

    def insert(self, h5file, description):
        """ Insert the data into an h5py File. """
        group = h5file.create_group(self.label)
        set_encoded(
            group.attrs,
            Description=description,
        )
        self.insert_into_group(group)

    def insert_into_group(self, group):
        """ Insert at the group level """
        self.prepare_data()
        self.record_group_attributes(group.attrs)
        for inserter in self:
            if isinstance(inserter, CompositeRecordInserter):
                # Sub-composites get their own new groups
                sub_group = group.create_group(inserter.label)
                inserter.insert_into_group(sub_group)
            else:
                # Simple data inserts below the composite group
                inserter.prepare_data()
                inserter.insert_below_group(group)



import numpy as np

from .record_inserter import inserter
from .numeric_inserter import ArrayInserter as NumericArrayInserter
from .utils import set_encoded


@inserter
class FileInserter(NumericArrayInserter):
    """ Inserter for file-like objects.

    File groups register as 'file' type, whereas the data registers as
    'numeric'.

    """

    record_type = 'file'

    def record_dataset_attributes(self, dict_like):
        attrs = dict(
            RecordType='numeric',
            Empty=self.empty,
            Complex='no',
            Sparse='no',
        )
        set_encoded(dict_like, **attrs)

    @staticmethod
    def can_insert(data):
        """ Can insert file-like objects. """
        return hasattr(data, 'read')

    def prepare_data(self):
        contents = self.data.read()
        if not isinstance(contents, bytes):
            contents = contents.encode('ascii')
        self.data = np.frombuffer(contents, dtype=np.uint8)
        NumericArrayInserter.prepare_data(self)



import numpy as np

from .record_inserter import CompositeRecordInserter, inserter
from .utils import cell_label, set_encoded


class CellRecordInserter(CompositeRecordInserter):
    """ Base class for cell record inserters. """

    record_type = 'cell'

    def record_group_attributes(self, dict_like):
        set_encoded(
            dict_like,
            RecordType=self.record_type,
            Empty=self.empty,
            Deflate=self.deflate,
            RecordSize=self.record_size,
        )


@inserter
class ListInserter(CellRecordInserter):
    """ Inserter for lists and tuples. """

    @staticmethod
    def can_insert(data):
        """ This can insert any list or tuple regardless of the contents. """
        return isinstance(data, (list, tuple))

    def prepare_data(self):
        """ Records RecordSize metadata. """
        self.empty = 'yes' if len(self.data) == 0 else 'no'
        self.record_size = (1, len(self.data))

    def __iter__(self):
        for i, sub_data in enumerate(self.data, start=1):
            cls = self.registry.get_inserter(sub_data)
            if cls is None:
                msg = "Data not supported for insertion: {!r}".format(sub_data)
                raise ValueError(msg)
            label = cell_label(i)
            inserter = cls(
                label, sub_data, self.deflate, registry=self.registry,
            )
            yield inserter


@inserter
class ArrayInserter(ListInserter):
    """ Inserter for arrays of objects. """

    @staticmethod
    def can_insert(data):
        """ This can insert arrays of objects or string-like things. """
        if not isinstance(data, np.ndarray):
            return False
        dtypes = (np.object_, np.unicode_, np.str_)
        return issubclass(data.dtype.type, dtypes)

    def prepare_data(self):
        """ Records RecordSize and Empty metadata. """
        self.record_size = np.atleast_2d(self.data).shape
        self.data = self.original_data.ravel(order='F')
        self.empty = 'yes' if self.data.size == 0 else 'no'



import numpy as np
from scipy.sparse import coo_matrix

from .utils import (
    cell_label, get_decoded, get_empty_for_type, is_simple, is_supported,
)


def extract(h5file, label):
    """ Extract data from an archive.

    Parameters
    ----------
    h5file : h5py.File
        The h5py File containing data
    label : str
        The data label.

    Returns
    -------
    data : object
        Archive data associated with the label.

    """
    grp = h5file[label]
    attrs = get_decoded(grp.attrs)
    return _extract_data_from_group(grp, label, attrs)


def extract_simple(record_type, data, data_attrs):
    """ Extract simple data from its raw storage format.

    Parameters
    -----------
    data : ndarray
        Data extracted from hdf5 dataset storage
    record_type : str
        The simple data type
    data_attrs : dict
        Attributes associated with the stored dataset

    Returns
    -------
    extracted :
        The extracted simple data

    """
    complex_flag = data_attrs.get('Complex', 'no')
    sparse_flag = data_attrs.get('Sparse', 'no')
    shape = data_attrs.get('ArraySize', None)

    if record_type == 'numeric':
        if sparse_flag == 'yes':
            if complex_flag == 'yes':
                extracted = extract_sparse_complex(data, shape.astype(int))
            else:
                extracted = extract_sparse(data)
        elif complex_flag == 'yes':
            extracted = extract_complex(data, shape.astype(int))
        else:
            extracted = extract_numeric(data)
    elif record_type == 'logical':
        extracted = extract_logical(data)
    elif record_type == 'character':
        extracted = extract_character(data)
    elif record_type == 'file':
        extracted = extract_file(data)

    return extracted


def extract_character(data):
    """ Extract 'character' data from uint8 stored form.

    Parameters
    -----------
    data : ndarray
        Array of uint8 ascii encodings

    Returns
    -------
    extracted : str or ndarray
        Reconstructed ascii string. If the input is a row array, then it will
        be converted to a string. Otherwise, it will be returned as an array of
        'S1' dtype.

    """
    # Matlab stores the transpose of 2D arrays. This must be unapplied here.
    data = data.T
    if data.ndim == 2 and data.shape[0] == 1:
        data = data.tobytes().decode('ascii')
    else:
        data = data.view('S1')
    return data


def extract_complex(data, shape):
    """ Extract complex 'numeric' data.

    Parameters
    -----------
    data : ndarray
        2 x N array containing real and imaginary portions of the complex data.
    shape : tuple
        Shape of the extracted array.

    Returns
    -------
    extracted : ndarray
        The extracted complex array.

    """
    extracted = 1j * data[1]
    extracted.real = data[0]
    extracted = extracted.reshape(shape, order='F')
    return reduce_array(extracted)


def extract_file(data):
    """ Extract file data.

    Parameters
    -----------
    data : ndarray
        The file contents as a byte array.

    Returns
    -------
    extracted : bytes
        The data is left alone.

    """
    return data.tobytes()


def extract_logical(data):
    """ Extract 'logical' data from uint8 stored form.

    Parameters
    -----------
    data : ndarray
        Array of uint8 values clipped to 0 or 1

    Returns
    -------
    extracted :
        The extracted boolean or boolean array

    """
    data = np.asarray(data, dtype=bool)
    return reduce_array(data.T)


def extract_numeric(data):
    """ Extract 'numeric' data from stored form.

    Parameters
    -----------
    data : ndarray or scalar
        Array or scalar of numeric data

    Returns
    -------
    data : ndarray or scalar
        The input data

    """
    # Matlab stores the transpose of 2D arrays. This must be unapplied here.
    return reduce_array(data.T)


def extract_sparse(data):
    """ Extract sparse 'numeric' data from stored form.

    Parameters
    -----------
    data : 3xN ndarray
        3xN array containing the rows, columns, and values of a sparse matrix
        in COO form. Note that the row and column arrays must be 1-based to be
        compatible with MATLAB.

    Returns
    -------
    extracted : scipy.sparse.coo_matrix
        The extracted sparse matrix

    """
    row, col, data = data
    # Fix 1-based indexing
    row -= 1
    col -= 1
    return coo_matrix((data, (row, col)))


def extract_sparse_complex(data, shape):
    """ Extract sparse 'numeric' data from stored form.

    Parameters
    -----------
    data : ndarray
        3xN array containing the index, real, and imaginary values of a
        sparse complex data. The index is unraveled and 1-based.
    shape : tuple
        Shape of the extracted array

    Returns
    -------
    extracted : coo_matrix
        The extracted sparse, complex matrix

    """
    index = data[0].astype(np.int64)
    # Fix 1-based indexing
    index -= 1
    data = extract_complex(data[1:], (data.shape[1],))
    row, col = np.unravel_index(index, shape)
    return coo_matrix((data, (row, col)))


def _extract_data_from_group(grp, label, attrs):
    """ Extract data from h5 group. ``label`` is the group label. """

    record_type = attrs['RecordType']
    if not is_supported(record_type):
        msg = "RecordType '{}' is not supported".format(record_type)
        raise ValueError(msg)

    empty = attrs['Empty']
    if empty == 'yes':
        return get_empty_for_type(record_type)

    if is_simple(record_type):
        ds = grp[label]
        data_attrs = get_decoded(ds.attrs)
        return extract_simple(record_type, ds[()], data_attrs)

    if record_type in ('cell', 'structures', 'objects'):
        record_size = attrs['RecordSize'].astype(int)
        nr = np.prod(record_size)
        labels = [cell_label(i) for i in range(1, nr + 1)]
        data = _extract_composite_data(grp, labels)
        if record_size[0] > 1 or len(record_size) > 2:
            data = np.array(
                data, dtype=object,
            ).reshape(record_size, order='F')
    elif record_type in ('structure', 'object'):
        labels = attrs['FieldNames'].split()
        data = _extract_composite_data(grp, labels)
        data = dict(zip(labels, data))
    return data


def _extract_composite_data(grp, labels):
    """ Extract composite data from a Group object with given labels. """
    extracted = []
    for label in labels:
        sub_obj = grp[label]
        attrs = get_decoded(sub_obj.attrs)
        record_type = attrs['RecordType']
        if is_simple(record_type):
            element = extract_simple(record_type, sub_obj[()], attrs)
        else:
            element = _extract_data_from_group(sub_obj, label, attrs)
        extracted.append(element)
    return extracted


def reduce_array(arr):
    """ Reduce a 2d row-array or scalar to 1 or 0 dimensions, respectively. """
    # squeeze leading dimension if this is a MATLAB row array
    if arr.ndim == 2 and arr.shape[0] == 1:
        # if it's a scalar, go all the way
        if arr.shape[1] == 1:
            arr = arr[0, 0]
        else:
            arr = np.squeeze(arr, axis=0)
    return arr



import numpy as np
from scipy.sparse import issparse

from .record_inserter import SimpleRecordInserter, inserter
from .utils import UNSUPPORTED_NUMERIC_TYPE_CODES, set_encoded


class BaseNumericInserter(SimpleRecordInserter):
    """ Base inserter for numeric types. """

    record_type = 'numeric'

    def record_dataset_attributes(self, dict_like):
        attrs = dict(
            RecordType=self.record_type,
            Empty=self.empty,
            Complex=self.complex,
            Sparse=self.sparse,
        )
        if self.complex == 'yes':
            attrs['ArraySize'] = self.array_size
        set_encoded(dict_like, **attrs)


@inserter
class ArrayInserter(BaseNumericInserter):
    """ Inserter for numeric arrays. """

    @staticmethod
    def can_insert(data):
        """ Can insert numeric arrays. """
        if not isinstance(data, np.ndarray):
            return False
        if data.dtype.char in UNSUPPORTED_NUMERIC_TYPE_CODES:
            return False
        return np.issubdtype(data.dtype, np.number)

    def prepare_data(self):
        self.empty = 'no'
        if np.iscomplexobj(self.data):
            # This prepares the data as a 2xN array containing the real and
            # imaginary values of the input as rows 0 and 1.
            self.complex = 'yes'
            data = np.atleast_2d(self.data)
            self.array_size = data.shape
            data = data.ravel(order='F')
            self.data = np.array([data.real, data.imag]).reshape(2, -1)
            if self.array_size == (1, 1) and np.all(np.isnan(self.data)):
                self.empty = 'yes'
        else:
            self.complex = 'no'
            self.data = np.atleast_2d(self.data).T
            self.array_size = None
            if self.data.shape == (1, 1) and np.all(np.isnan(self.data)):
                self.empty = 'yes'

        self.sparse = 'no'
        self.empty = 'yes' if self.data.size == 0 else self.empty


@inserter
class ScalarInserter(ArrayInserter):
    """ Inserter for numeric scalars. """

    @staticmethod
    def can_insert(data):
        """ Can insert numeric scalars. """
        types = (float, complex, int, np.long)
        if isinstance(data, types) and not isinstance(data, bool):
            return True
        elif isinstance(data, np.number):
            return data.dtype.char not in UNSUPPORTED_NUMERIC_TYPE_CODES

    def prepare_data(self):
        """ Records RecordSize metadata. """
        self.data = np.asarray(self.data)
        ArrayInserter.prepare_data(self)


@inserter
class SparseInserter(BaseNumericInserter):
    """ Inserter for sparse, numeric arrays. """

    @staticmethod
    def can_insert(data):
        """ Can insert numeric scalars. """
        if not issparse(data):
            return False
        if data.dtype.char in UNSUPPORTED_NUMERIC_TYPE_CODES:
            return False
        return np.issubdtype(data.dtype, np.number)

    def prepare_data(self):
        """ Records RecordSize metadata. """
        self.sparse = 'yes'
        data = self.data.tocoo()
        if np.issubdtype(self.data.dtype, np.complexfloating):
            self.complex = 'yes'
            self.array_size = data.shape
            indices = np.ravel_multi_index((data.row, data.col), data.shape)
            self.data = np.vstack([
                indices + 1, data.data.real, data.data.imag
            ])
        else:
            self.complex = 'no'
            self.array_size = None
            # 3 x N, [row, column, value], 1-based
            self.data = np.array([data.row + 1, data.col + 1, data.data])



import numpy as np

from .record_inserter import SimpleRecordInserter, inserter


@inserter
class ArrayInserter(SimpleRecordInserter):
    """ Inserter for boolean arrays. """

    record_type = 'logical'

    @staticmethod
    def can_insert(data):
        """ Can insert boolean arrays. """
        if not isinstance(data, np.ndarray):
            return False
        return np.issubdtype(data.dtype, np.bool_)

    def prepare_data(self):
        data = self.data.astype(np.uint8).clip(0, 1)
        # Matlab stores the transpose of 2D arrays. This must be applied here.
        self.data = np.atleast_2d(data).T
        self.empty = 'yes' if self.data.size == 0 else 'no'


@inserter
class ScalarInserter(ArrayInserter):
    """ Inserter for bool and np.bool_. """

    @staticmethod
    def can_insert(data):
        """ Can insert bool and np.bool_ """
        return isinstance(data, (bool, np.bool_))

    def prepare_data(self):
        """ Records RecordSize metadata. """
        self.data = np.asarray(self.data)
        ArrayInserter.prepare_data(self)



from contextlib import contextmanager
import os
import os.path as op
import tempfile
import unittest

import h5py
import numpy as np
from numpy.testing import assert_equal
from scipy.sparse import coo_matrix, eye

import sdafile
from sdafile.utils import get_decoded


DATA_DIR = op.join(op.abspath(op.dirname(sdafile.__file__)), 'tests', 'data')


def data_path(fname):
    """ Get path to data in test data directory. """
    return op.join(DATA_DIR, fname)


BAD_ATTRS = {
    'FileFormat': 'SDB',
    'FormatVersion': '0.5',
    'Writable': 'nope',
    'Created': '2017-01-01 01:23:45',
    'Updated': '2017-01-01 01:23:45',
}


GOOD_ATTRS = {
    'FileFormat': 'SDA',
    'FormatVersion': '1.1',
    'Writable': 'yes',
    'Created': '18-Aug-2017 01:23:45',
    'Updated': '18-Aug-2017 01:23:45',
}


FLOAT_VAL = 3.14159
INT_VAL = 3
BOOL_VAL = True
COMPLEX_VAL = 1.23 + 4.56j
STR_VAL = 'foo'
UNICODE_VAL = u'foo'

TEST_NUMERIC = [
    FLOAT_VAL,
    np.float32(FLOAT_VAL),
    np.float64(FLOAT_VAL),
    INT_VAL,
    np.long(INT_VAL),
    np.int8(INT_VAL),
    np.int16(INT_VAL),
    np.int32(INT_VAL),
    np.int64(INT_VAL),
    np.uint8(INT_VAL),
    np.uint16(INT_VAL),
    np.uint32(INT_VAL),
    np.uint64(INT_VAL),
    COMPLEX_VAL,
    np.complex64(COMPLEX_VAL),
    np.complex128(COMPLEX_VAL),
]

TEST_NUMERIC += [
    np.array(val) for val in TEST_NUMERIC
] + [
    np.array([val] * 4) for val in TEST_NUMERIC
] + [
    np.array([val] * 6).reshape(2, 3) for val in TEST_NUMERIC
]

TEST_LOGICAL = [
    BOOL_VAL,
    np.bool_(BOOL_VAL),
]

TEST_LOGICAL += [
    np.array(val) for val in TEST_LOGICAL
] + [
    np.array([val] * 4) for val in TEST_LOGICAL
] + [
    np.array([val] * 6).reshape(2, 3) for val in TEST_LOGICAL
]

TEST_CHARACTER = [
    STR_VAL,
    np.str_(STR_VAL),
    np.unicode_(STR_VAL),
]

TEST_CHARACTER += [
    np.array(list(val)).reshape(-1, 1) for val in TEST_CHARACTER
]


# Sparse matrix in all forms
TEST_SPARSE = [coo_matrix((np.arange(5), (np.arange(1, 6), np.arange(2, 7))))]
TEST_SPARSE.extend([
    TEST_SPARSE[0].tocsr(), TEST_SPARSE[0].tocsc(), TEST_SPARSE[0].tolil(),
    TEST_SPARSE[0].tobsr(), TEST_SPARSE[0].todok()
])

TEST_SPARSE_COMPLEX = [
    coo_matrix((np.arange(5) * (1 + 2j), (np.arange(1, 6), np.arange(2, 7))))
]
TEST_SPARSE_COMPLEX.extend([
    TEST_SPARSE_COMPLEX[0].tocsr(), TEST_SPARSE_COMPLEX[0].tocsc(),
    TEST_SPARSE_COMPLEX[0].tolil(), TEST_SPARSE_COMPLEX[0].tobsr(),
    TEST_SPARSE_COMPLEX[0].todok()
])

# lists, tuples
TEST_CELL = [
    ['hi', 'hello'],
    np.array(['hi', 'hello']),
    ['hello', np.arange(4)],
    ['hello', [True, np.arange(4)]],
    ['hello', (True, np.arange(4))],
    np.array(['hello', 3, [True, False, True], 3.14], dtype=object),
    np.array(
        [
            ['hello', 3],
            [[True, False, True], 3.14]
        ],
        dtype=object
    ),
    np.array(
        [
            ['hello', 3],
            [[True, False, True], 3.14]
        ],
        dtype=object,
        order='F',
    ),
    [
        {
            'foo': 'foo',
            'bar': np.arange(4),
            'baz': np.array([True, False])
        },
    ] * 3,
]

TEST_STRUCTURE = [
    {
        'foo': 'foo',
        'bar': np.arange(4),
        'baz': np.array([True, False])
    },
    {
        'foo': 'foo',
        'bar': [np.arange(4), np.array([True, False])]
    },
    {
        'strings': ['hi', 'hello'],
        'structure': {
            'foo': 'foo',
            'bar': np.arange(4),
            'baz': np.array([True, False])
        }
    },
]


# Unsupported
TEST_UNSUPPORTED = [
    eye(5, dtype=bool),  # sparse bool
    lambda x: x**2,
    {0},
    None,
]


# unsupported types, platform-specific
if hasattr(np, 'complex256'):
    TEST_UNSUPPORTED.append(np.complex256(0))
    TEST_UNSUPPORTED.append(np.arange(5, dtype=np.complex256))
    TEST_UNSUPPORTED.append(eye(5, dtype=np.complex256))
if hasattr(np, 'float128'):
    TEST_UNSUPPORTED.append(np.float128(0))
    TEST_UNSUPPORTED.append(np.arange(5, dtype=np.float128))
    TEST_UNSUPPORTED.append(eye(5, dtype=np.float128))
if hasattr(np, 'float16'):
    TEST_UNSUPPORTED.append(np.float16(0))
    TEST_UNSUPPORTED.append(np.arange(5, dtype=np.float16))
    TEST_UNSUPPORTED.append(eye(5, dtype=np.float16))


@contextmanager
def temporary_file(suffix='.sda'):
    pid, file_path = tempfile.mkstemp(suffix=suffix)
    os.close(pid)
    try:
        yield file_path
    finally:
        if op.isfile(file_path):
            os.remove(file_path)


@contextmanager
def temporary_h5file(suffix='.sda'):
    with temporary_file(suffix) as file_path:
        h5file = h5py.File(file_path, 'w')
        try:
            yield h5file
        finally:
            if h5file.id.valid:  # file is open
                h5file.close()


class MockRecordInserter(object):
    """ RecordInserter for testing.

    This must be used instantiated.

    """

    record_type = 'testing'

    def __init__(self, called):
        self.called = called

    def __call__(self, label, data, deflate, registry=None):
        # Mock initialization.
        self.label = label
        self.deflate = int(deflate)
        self.data = self.original_data = data
        self.empty = 'no'
        self._registry = registry
        return self

    def can_insert(self, data):
        return True

    def insert(self, h5file, description):
        self.called.append(description)


class InserterTestCase(unittest.TestCase):

    def setUp(self):
        from sdafile.record_inserter import InserterRegistry
        self.registry = InserterRegistry()

    def tearDown(self):
        del self.registry

    @contextmanager
    def insert(self, cls, label, data, deflate, description):
        inserter = cls(label, data, deflate, self.registry)
        with temporary_h5file() as h5file:
            inserter.insert(h5file, description)
            yield h5file

    def assertAttrs(self, dict_like, **attrs):
        assert_equal(attrs, get_decoded(dict_like))

    def assertRegistry(self, cls, data):
        """ Assert registry works for data. """
        self.assertTrue(cls.can_insert(data))
        found_cls = self.registry.get_inserter(data)
        self.assertIs(found_cls, cls)

    def assertSimpleInsert(self, cls, data, group_attrs, ds_attrs, expected):
        """ Test simple insertion. Pass expected=None to skip data check. """
        # Check registration
        self.assertRegistry(cls, data)

        # Test insertion
        label = 'test'
        with self.insert(cls, label, data, 0, 'desc') as h5file:
            grp = h5file[label]
            self.assertAttrs(
                grp.attrs,
                Description='desc',
                Deflate=0,
                **group_attrs
            )
            ds = grp[label]
            self.assertAttrs(ds.attrs, **ds_attrs)
            if expected is not None:
                stored = ds[()]
                assert_equal(stored, expected)



import numpy as np

from .record_inserter import SimpleRecordInserter, inserter


@inserter
class ArrayInserter(SimpleRecordInserter):
    """ Inserter for character arrays. """

    record_type = 'character'

    @staticmethod
    def can_insert(data):
        """ Insert boolean arrays. """
        if not isinstance(data, np.ndarray):
            return False
        return data.dtype == np.dtype('S1')

    def prepare_data(self):
        data = self.data.view(np.uint8)
        # Matlab stores the transpose of 2D arrays. This must be applied here.
        self.data = np.atleast_2d(data).T
        self.empty = 'yes' if self.data.size == 0 else 'no'


@inserter
class StringInserter(ArrayInserter):
    """ Inserter for strings. """

    @staticmethod
    def can_insert(data):
        """ Insert bool and np.bool_ """
        return isinstance(data, (str, np.unicode))

    def prepare_data(self):
        self.data = np.frombuffer(self.data.encode('ascii'), 'S1')
        ArrayInserter.prepare_data(self)


if bytes is str:  # python 3

    BytesInserter = StringInserter

else:  # python 2

    @inserter
    class BytesInserter(ArrayInserter):
        """ Inserter for bytes. """

        @staticmethod
        def can_insert(data):
            """ Insert bytes. """
            return isinstance(data, bytes)

        def prepare_data(self):
            self.data = np.frombuffer(self.data, 'S1')
            ArrayInserter.prepare_data(self)



import os

from setuptools import setup, find_packages

# Get version from version.py.
version_filename = os.path.join("sdafile", "version.py")
with open(version_filename) as version_file:
    version_py_locals = {}
    exec(version_file.read(), version_py_locals)
    version = version_py_locals["version"]


install_requires = [
    "h5py", 
    "scipy", 
    "pandas", 
    "numpy", 
    "packaging", 
]


setup(
    name="sdafile",
    version=version,
    packages=find_packages(),
    author="Christopher L. Farrow",
    author_email="cfarrow@enthought.com",
    maintainer="Sandia National Laboratories",
    url="https://github.com/enthought/sandia-data-archive",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: OS Independent",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
    ],
    description=(
        'Tools for reading, writing, altering, and inspecting Sandia Data '
        'Archive (SDA) files.'
    ),
    install_requires=install_requires,
    license='BSD',
    platforms=["Windows", "Linux", "Mac OS-X", "Unix", "Solaris"],
    zip_safe=True,
)



""" Implementation of ``SDAFile`` for working with SDA files.

The SDA format was designed to be universal to facilitate data sharing across
multiple languages. It supports reading and updating all record types, except
*function* records. It support writing *numeric*, *logical*, *cell*, and
*structure* records.

"""

from contextlib import contextmanager
from functools import partial
import os
import os.path as op
import re
import shutil
import tempfile

import h5py
import numpy as np

from .extract import extract
from .record_inserter import InserterRegistry
from .utils import (
    are_signatures_equivalent, error_if_bad_header, error_if_not_writable,
    get_decoded, is_valid_writable, set_encoded, unnest, unnest_record,
    update_header, validate_structures, write_header,
)


WRITE_MODES = ('w', 'w-', 'x', 'a')


class SDAFile(object):
    """ Read, write, inspect, and manipulate Sandia Data Archive files.

    This supports version 1.1 of the Sandia Data Archive format.

    """

    def __init__(self, name, mode='a', **kw):
        """ Open an SDA file for reading, writing, or interrogation.

        Parameters
        ----------
        name : str
            The name of the file to be loaded or created.
        mode : str
            r         Read-only, file must exist
            r+        Read/write, file must exist
            w         Create file, truncate if exists
            w- or x   Create file, fail if exists
            a         Read/write if exists, create otherwise (default)
        kw :
            Key-word arguments that are passed to the underlying HDF5 file. See
            h5py.File for options.

        """
        file_exists = op.isfile(name)
        self._mode = mode
        self._filename = name
        self._kw = kw
        self._registry = InserterRegistry()

        # Check existence
        if mode in ('r', 'r+') and not file_exists:
            msg = "File '{}' does not exist".format(name)
            raise IOError(msg)

        # Check the header when mode requires the file to exist
        if mode in ('r', 'r+') or (file_exists and mode == 'a'):
            with self._h5file('r') as h5file:
                error_if_bad_header(h5file)

            # Check that file is writable when mode will write to file.
            if mode != 'r':
                with self._h5file('a') as h5file:
                    error_if_not_writable(h5file)

        # Create the header if this is a new file
        if mode in ('w', 'w-', 'x') or (not file_exists and mode == 'a'):
            with self._h5file(mode) as h5file:
                write_header(h5file.attrs)

    # File properties

    @property
    def name(self):
        """ File name on disk. """
        return self._filename

    @property
    def mode(self):
        """ Mode used to open file. """
        return self._mode

    # Format attrs

    @property
    def FileFormat(self):
        """ The 'FileFormat' file attribute. """
        return self._get_attr('FileFormat')

    @property
    def FormatVersion(self):
        """ The format version from the SDA file. """
        return self._get_attr('FormatVersion')

    @property
    def Writable(self):
        """ The 'Writable' flag from the SDA file. """
        return self._get_attr('Writable')

    @Writable.setter
    def Writable(self, value):
        if self._mode not in WRITE_MODES:
            raise ValueError("File is not writable.")
        if not is_valid_writable(value):
            raise ValueError("Must be 'yes' or 'no'")
        with self._h5file('r+') as h5file:
            set_encoded(h5file.attrs, Writable=value)

    @property
    def Created(self):
        """ The time the file was created. """
        return self._get_attr('Created')

    @property
    def Updated(self):
        """ The time the file was last updated. """
        return self._get_attr('Updated')

    # Public
    def describe(self, label, description=''):
        """ Change the description of a data entry.

        Parameters
        ----------
        label : str
            The data label.
        description : str
            A description to accompany the data

        Raises
        ------
        ValueError if the label contains invalid characters
        ValueError if the label does not exist

        """
        self._validate_can_write()
        self._validate_labels(label, must_exist=True)
        with self._h5file('r+') as h5file:
            set_encoded(h5file[label].attrs, Description=description)
            update_header(h5file.attrs)

    def extract(self, label):
        """ Extract data from an SDA file.

        Parameters
        ----------
        label : str
            The data label.

        Returns
        -------
        data : object
            Archive data associated with the label.

        Notes
        -----
        Sparse numeric data is extracted as
        :class:`coo_matrix<scipy:scipy.sparse.coo_matrix>`. This format does
        not support all numpy operations.

        Raises
        ------
        ValueError if the label contains invalid characters
        ValueError if the label does not exist

        """
        self._validate_labels(label, must_exist=True)
        with self._h5file('r') as h5file:
            return extract(h5file, label)

    def extract_to_file(self, label, path, overwrite=False):
        """ Extract a file record to file.

        Parameters
        ----------
        label : str
            Label of the file record.
        path : str
            The path to which the file is to be written.
        overwrite : bool, optional
            Unless specified as True, an existing file with the chosen name
            will not be overwritten by this method.

        Raises
        ------
        IOError if `overwrite` is False and the destintation file exists.

        """

        if op.exists(path) and not overwrite:
            raise IOError("File '{}' exists. Will not overwrite.".format(path))
        self._validate_labels(label, must_exist=True)

        # Check that archive is a file archive
        record_type = self._get_attr('RecordType', root=label)
        if record_type != 'file':
            raise ValueError("'{}' is not a file record".format(label))

        with open(path, 'wb') as f:
            f.write(self.extract(label))

    def insert(self, label, data, description='', deflate=0,
               as_structures=False):
        """ Insert data into an SDA file.

        Parameters
        ----------
        label : str
            The data label.
        data :
            The data to insert. See the notes below.
        description : str, optional
            A description to accompany the data
        deflate : int, optional
            An integer value from 0 to 9, specifying the compression level to
            be applied to the stored data.
        as_record : bool, optional
            If specified, data that is storable as a cell record and has
            homogenous cells will be stored as a "structures" record. Note that
            this does not extend to nested cell records.

        Raises
        ------
        ValueError if the data is of an unsupported type
        ValueError if the label contains invalid characters
        ValueError if the label exists
        ValueError if `as_structures` is True and the data cannot be stored as
        a structures record.

        Notes
        -----
        This stores specific data types as described here.

        sequences :
            Lists, tuples, and anything else that identifies as a
            collections.abc.Sequence are stored as 'cell' records, no matter
            the contents.

        dicts :
            Dictionaries are stored as 'structure' records.

        numpy arrays :
            If the dtype is a supported numeric type, then a numpy array is
            stored as a 'numeric' record. Arrays of 'bool' type are stored as
            'logical' records.  Arrays of characters (dtype 'S1') are stored as
            'character' records. Object and string arrays are stored as 'cell'
            records.

        sparse arrays (:class:`coo_matrix<scipy:scipy.sparse.coo_matrix>`) :
            These are stored as 'numeric' records if the dtype is a type
            supported for numeric numpy arrays.

        strings :
            Strings are stored as 'character' records. An attempt will be
            made to convert the input to ascii encoded bytes, no matter the
            underlying encoding. This may result in an encoding exception if
            the input cannot be ascii encoded.

        non-string scalars :
            Non-string scalars are stored as 'numeric' if numeric, or 'logical'
            if boolean.

        file-like :
            The contents of a file-like objects (with a 'read' method) are
            stored as 'file' records.

        other :
            Arrays of characters are not supported. Convert to a string.
            Object arrays are not supported. Cast to another dtype or turn into
            a list.

        Anything not listed above is not (intentionally) supported.

        See Also
        --------
        insert_from_file : Insert contents of a named file.

        """
        self._validate_can_write()
        self._validate_labels(label, can_exist=False)
        if not isinstance(deflate, (int, np.integer)) or not 0 <= deflate <= 9:
            msg = "'deflate' must be an integer from 0 to 9"
            raise ValueError(msg)
        cls = self._registry.get_inserter(data)
        if cls is None:
            msg = "{!r} is not a supported type".format(data)
            raise ValueError(msg)

        inserter = cls(label, data, deflate, self._registry)

        if as_structures:
            if inserter.record_type != 'cell':
                msg = "Data cannot be stored as a 'structures' record."
                raise ValueError(msg)

            validate_structures(data, self._registry)

            # Tell the inserter to use the 'structures' record type
            inserter.record_type = 'structures'

        with self._h5file('r+') as h5file:
            try:
                inserter.insert(h5file, description)
            except Exception:
                # Do not leave things in an invalid state
                if label in h5file:
                    del h5file[label]
                raise
            else:
                update_header(h5file.attrs)

    def insert_from_file(self, path, description='', deflate=0):
        """ Insert the contents of a file as a file record.

        Parameters
        ----------
        path : str
            The path to the file. The basename of the path will be used as the
            record label.
        description : str, optional
            A description to accompany the data
        deflate : int, optional
            An integer value from 0 to 9, specifying the compression level to
            be applied to the stored data.

        Returns
        -------
        label : str
            The label under which the file was stored.

        See Also
        --------
        insert : Insert data into the archive

        """
        if not op.isfile(path):
            raise ValueError("File '{}' does not exist".format(path))

        label = op.basename(path)
        with open(path, 'rb') as f:
            self.insert(label, f, description, deflate)
        return label

    def labels(self):
        """ Get data labels from the archive.

        Returns
        -------
        labels : list of str
            Labels from the archive.

        """
        with self._h5file('r') as h5file:
            return list(h5file.keys())

    def remove(self, *labels):
        """ Remove specified labels from the archive.

        This cannot be undone.

        """
        self._validate_can_write()
        self._validate_labels(labels, must_exist=True)

        # Create a new file so space is actually freed
        def _copy_visitor(path, source, destination, labels):
            """ Visitor that copies data from source to destination """

            # Skip paths corresponding to excluded labels
            if path.split('/')[0] in labels:
                return

            # Copy everything else
            source_obj = source[path]
            if isinstance(source_obj, h5py.Group):
                dest_obj = destination.create_group(path)
            else:
                ds = source_obj
                dest_obj = destination.create_dataset(
                    path,
                    data=source_obj[()],
                    chunks=ds.chunks,
                    maxshape=ds.maxshape,
                    compression=ds.compression,
                    compression_opts=ds.compression_opts,
                    scaleoffset=ds.scaleoffset,
                    shuffle=ds.shuffle,
                    fletcher32=ds.fletcher32,
                    fillvalue=ds.fillvalue,
                )

            dest_obj.attrs.update(source_obj.attrs)

        pid, destination_path = tempfile.mkstemp()
        os.close(pid)
        with h5py.File(destination_path, 'w') as destination:
            with self._h5file('r') as source:
                destination.attrs.update(source.attrs)
                source.visit(
                    partial(
                        _copy_visitor,
                        source=source,
                        destination=destination,
                        labels=set(labels),

                    )
                )
            update_header(destination.attrs)
        shutil.move(destination_path, self._filename)

    def probe(self, pattern=None):
        """ Summarize the state of the archive

        This requires the pandas package.

        Parameters
        ----------
        pattern : str or None, optional
            A search pattern (python regular expression) applied to find
            archive labels of interest. If None, all labels are selected.

        Returns
        -------
        summary : :class:`DataFrame<pandas:pandas.DataFrame>`
            A table summarizing the archive.

        """
        from pandas import DataFrame
        labels = self.labels()
        if pattern is not None:
            regex = re.compile(pattern)
            labels = [
                label for label in labels if regex.match(label) is not None
            ]

        summary = []
        with self._h5file('r') as h5file:
            for label in labels:
                g = h5file[label]
                attrs = get_decoded(g.attrs)
                if label in g:
                    attrs.update(get_decoded(g[label].attrs))
                attrs['label'] = label
                summary.append(attrs)

        cols = [
            'label', 'RecordType', 'Description', 'Empty', 'Deflate',
            'Complex', 'ArraySize', 'Sparse', 'RecordSize', 'Class',
            'FieldNames', 'Command',
        ]
        return DataFrame(summary, columns=cols).set_index('label').fillna('')

    def replace(self, label, data):
        """ Replace an existing dataset.

        Parameters
        ----------
        label : str
            The record label.
        data :
            The data with which to replace the record.

        Notes
        -----
        This is equivalent to removing the data and inserting a new entry using
        the same ``label``, ``description``, and ``deflate`` options.

        """
        self._validate_can_write()
        self._validate_labels(label, must_exist=True)
        with self._h5file('r+') as h5file:
            attrs = get_decoded(h5file[label].attrs, 'Deflate', 'Description')
            del h5file[label]
        self.insert(label, data, attrs['Description'], attrs['Deflate'])

    def update_object(self, label, data):
        """ Update an existing object record.

        Parameters
        ----------
        label : str
            Label of the object record.
        data : dict
            The data with which to replace the object record.

        Notes
        -----
        This is more strict than **replace** in that the intention is to update
        the contents of an 'object' record while preserving the record type.
        The simplest way to make use of this is to *extract* an object record,
        replace some data, and then call this to update the stored record.

        """
        self._validate_can_write()
        self._validate_labels(label, must_exist=True)

        cls = self._registry.get_inserter(data)
        if cls is None:
            msg = "{!r} is not a supported type".format(data)
            raise ValueError(msg)

        record_type = cls.record_type
        if record_type != 'structure':
            raise ValueError("Input data is not a dictionary")

        with self._h5file('r+') as h5file:
            # Check the general structure of the data and file
            grp = h5file[label]
            attrs = get_decoded(grp.attrs)
            if not attrs['RecordType'] == 'object':
                raise ValueError("Record '{}' is not an object".format(label))
            if attrs['Empty'] == 'yes':
                raise ValueError("Cannot update an empty record")
            record_sig = unnest_record(grp)
            data_sig = unnest(data, self._registry)
            if not are_signatures_equivalent(record_sig, data_sig):
                msg = "Data is not compatible with record '{}'"
                raise ValueError(msg.format(label))

            del h5file[label]

        self.insert(label, data, attrs['Description'], int(attrs['Deflate']))

        # Fix the record type and update the header
        with self._h5file('r+') as h5file:
            grp = h5file[label]
            set_encoded(
                grp.attrs,
                RecordType='object',
                Class=attrs['Class'],
            )
            update_header(h5file.attrs)

    def update_objects(self, label, data):
        """ Update an existing objects record.

        Parameters
        ----------
        label : str
            Label of the objects record.
        data : list
            The data with which to replace the objects record.

        Notes
        -----
        This is more strict than **replace** in that the intention is to update
        the contents of an 'objects' record while preserving the record type.
        The simplest way to make use of this is to *extract* an objects record,
        replace some data, and then call this to update the stored record.

        """
        self._validate_can_write()
        self._validate_labels(label, must_exist=True)

        cls = self._registry.get_inserter(data)
        if cls is None:
            msg = "{!r} is not a supported type".format(data)
            raise ValueError(msg)

        record_type = cls.record_type
        if record_type != 'cell':
            raise ValueError("Input data is not a list")

        # To be an 'objects' record, this must look like a 'structures' record.
        data_sig = validate_structures(data, self._registry)

        with self._h5file('r+') as h5file:
            # Check the general structure of the data and file
            grp = h5file[label]
            attrs = get_decoded(grp.attrs)
            if not attrs['RecordType'] == 'objects':
                raise ValueError("Record '{}' is not an objects".format(label))
            if attrs['Empty'] == 'yes':
                raise ValueError("Cannot update an empty record")
            record_sig = unnest_record(grp['element 1'])
            if not are_signatures_equivalent(record_sig, data_sig):
                msg = "Data is not compatible with record '{}'"
                raise ValueError(msg.format(label))

            del h5file[label]

        self.insert(label, data, attrs['Description'], int(attrs['Deflate']))

        # Fix the record type and update the header
        with self._h5file('r+') as h5file:
            grp = h5file[label]
            set_encoded(
                grp.attrs,
                RecordType='objects',
                Class=attrs['Class'],
            )
            update_header(h5file.attrs)

    # Private

    @contextmanager
    def _h5file(self, mode):
        h5file = h5py.File(self._filename, mode, **self._kw)
        try:
            yield h5file
        finally:
            h5file.close()

    def _get_attr(self, attr, root=None):
        """ Get a named atribute as a string """
        with self._h5file('r') as h5file:
            if root is None:
                obj = h5file
            else:
                obj = h5file[root]
            return get_decoded(obj.attrs, attr)[attr]

    def _validate_can_write(self):
        """ Validate file mode and 'Writable' attr allow writing. """
        if self._mode not in WRITE_MODES:
            raise IOError("File is not writable")
        if self.Writable == 'no':
            raise IOError("'Writable' flag is 'no'")

    def _validate_labels(self, labels, can_exist=True, must_exist=False):
        if isinstance(labels, str):
            labels = [labels]
        if len(labels) == 0:
            raise ValueError("Must specify labels")
        for label in labels:
            if '/' in label or '\\' in label:
                msg = r"label cannot contain '/' or '\'"
                raise ValueError(msg)
        with self._h5file('r') as h5file:
            for label in labels:
                label_exists = label in h5file
                if not can_exist and label_exists:
                    msg = "Label '{}' already exists.".format(label)
                    raise ValueError(msg)
                if must_exist and not label_exists:
                    msg = "Label item '{}' does not exist".format(label)
                    raise ValueError(msg)
